package piglets.gamma;

import piglets.SuperPiglet;
import piglets.SuperHerder;
import piglets.DrawWell;
import piglets.GladeCell;
import universum.bi.*;
import universum.util.Util;

/**
 * Created by IntelliJ IDEA.
 * User: Tanya
 * Date: 03.02.2007
 * Time: 12:20:57
 */

enum PigletState {
    EXPLORE,
    EATING,
    FEEDING,
    SURVIVE,
    ATTACK,
}

class EventException extends Exception {
    public final Event event;

    EventException(EventKind kind, Object arg) {
        this.event = new Event(kind, arg);
    }
}

public class Piglet extends SuperPiglet {
    public static final float M0                    = 5;
    public static final float S0                    = 3;

    public static final float C_crit                = Constants.K_emin + 5 * Constants.K_masscost;
    public static final float C_norm                = C_crit + 5 * Constants.K_masscost;
    public static final float C_growt_weight        = 5.f;
    public static final float C_energy_weight       = 0.05f;
    public static final float C_steps_to_born       = 0.3f / Constants.K_bite;
    public static final float C_max_steps_to_dw     = 2.f;

    public static final int C_growth_barrier        = 30;
    public static final int C_fight_barrier         = 1000;
    public static final int C_enemy_history         = 1;

    private PigletState state = PigletState.EXPLORE;

    private float E;
    private Location location;
    private PointInfo pointInfo;
    private BeingInterface beingInterface;

    public Event makeTurn(BeingInterface bi) {
        herder.updateTurn(bi);

        // update piglet params
        beingInterface = bi;
        E = bi.getEnergy(this);
        location = bi.getLocation(this);
        pointInfo = bi.getPointInfo(this);

        // exec automaton
        final int MAX_STEPS = 100;
        for (int step_ = 0; step_ < MAX_STEPS; ++step_) {
            PigletState prevState = state;

            try {
                switch (state) {
                    case EXPLORE:   onExplore();  break;
                    case EATING:    onEating();   break;
                    case FEEDING:   onFeeding();  break;
                    case SURVIVE:   onSurvive();  break;
                    case ATTACK:    onAttack();   break;
                }
            } catch (EventException event) {
                return event.event;
            }

            // break cycle if no action occurs
            if (state == prevState) {
                break;
            }
        }

        return null;
    }

    private void onExplore() throws EventException {
        /*
         	    .  
         */
        Location enemyPos = findClosestEnemy();
        float enemyDist = enemyPos != null ? distance(location, enemyPos) : Herder.INF;
        if (enemyDist < S + Herder.EPS) {
            log("EXPLORE: enemy found at " + enemyPos + "  -> ATTACK");
            state = PigletState.ATTACK;
            return ;
        }

        /*
         	 ,     ()
         */
        if (herder.hasEatingDrawWells() &&
            E > M * Constants.K_toborn &&
            (herder.MAX_TURNS - herder.getTurn()) * M * Constants.K_masscost >
            E * (0.5f - Constants.K_borncost - C_crit)
        ) {
            log("EXPLORE: born " + getParams());
            throw new EventException(EventKind.ACTION_BORN, getParams());
        }

        /*
         	  ,    (   )    
            (  ), .  
         */
        DrawWell drawWell = findClosestDrawWell();
        if (drawWell != null) {
            eatingDrawWell = drawWell;
            
            herder.lockWrite();
            drawWell.owners.add(this);
            herder.addEatingDrawWells(true);
            herder.unlockWrite();

            log("EXPLORE: eat at " + drawWell.pos + " -> EATING");
            
            state = PigletState.EATING;
            return ;
        }

        /*
         	  ,    , .  
         */
        Location next = findClosestFeeding();
        if (next != null) {
            feedingTarget = next;

            log("EXPLORE: feeding at " + feedingTarget + " -> FEEDING");

            state = PigletState.FEEDING;
            return ;
        }

        /*
         	  ,    . .        
         */
        if (E - M * C_crit < Constants.K_masscost * M * (herder.MAX_TURNS - herder.getTurn())) {
            state = PigletState.SURVIVE;

            log("EXPLORE: too few energy -> SURVIVE");
            return ;
        }

        /*
         	     
         */
        next = findExploreStep();
        if (next != null) {
            log("EXPLORE: move to " + next + " (near)");
            throw new EventException(EventKind.ACTION_MOVE_TO, next);
        }

        /*
         	  ,     
         */
        next = findClosestUnexploredCell();
        if (checkDistanceConstraints(next)) {
            log("EXPLORE: move to " + next + " (long)");
            throw new EventException(EventKind.ACTION_MOVE_TO, findStep(location, next));
        }

        /*
         	  
         */
        log("EXPLORE: otherwise -> SURVIVE");
        state = PigletState.SURVIVE;
    }

    private DrawWell eatingDrawWell;
    private void onEating() throws EventException {
        /*
         	       
         */
        if (distance(location, eatingDrawWell.pos) > Herder.EPS) {
            log("EATING: go to " + eatingDrawWell.pos);
            throw new EventException(EventKind.ACTION_MOVE_TO, findStep(location, eatingDrawWell.pos));
        }

        /*
         	   
         */
        if (E > M * Constants.K_toborn) {
            log("EATING: born " + getParams());
            throw new EventException(EventKind.ACTION_BORN, getParams());
        }

        /*
         	      (  )  .  
         */
        final float canEat = eatingDrawWell.getCell().count / eatingDrawWell.owners.size();
        if (eatingDrawWell.owners.size() > 1 && canEat < (M - E) / C_steps_to_born) {
            herder.lockWrite();
            eatingDrawWell.owners.remove(this);
            herder.unlockWrite();

            herder.addEatingDrawWells(false);
            eatingDrawWell = null;

            state = PigletState.SURVIVE;

            log("EATING: bad eating condition -> SURVIVE");
            return ;
        }

        /*
         	  ,    
         */
        log("EATING: eat " + (M - E) + "/" + canEat);
        throw new EventException(EventKind.ACTION_EAT, Math.min(canEat, M-E));
    }

    private Location feedingTarget;
    private void onFeeding() throws EventException {
        /*
         	        
         */
        if (distance(location, feedingTarget) > Herder.EPS) {
            log("FEEDING: go to " + feedingTarget);
            throw new EventException(EventKind.ACTION_MOVE_TO, findStep(location, feedingTarget));
        }

        /*
         	   
         */
        if (E > M * Constants.K_toborn) {
            log("FEEDING: born " + getParams());
            throw new EventException(EventKind.ACTION_BORN, getParams());
        }

        /*
         	 ,   
         */
        final float count = pointInfo.getCount(this);
        final float growth = pointInfo.getGrowthRate(this); 
        if (count < Herder.EPS || isGrowingPeriod() && (count + growth * C_steps_to_born < M * Constants.K_toborn - E)) {
            log("FEEDING: nothing to eat -> EXPLORE");
            state = PigletState.EXPLORE;
            return ;
        }

        /*
         	  
         */
        log("FEEDING: eat " + (M - E));
        throw new EventException(EventKind.ACTION_EAT, M - E);
    }

    private boolean tryNoToSurvive = false;
    private void onSurvive() throws EventException {
        /*
         	        (   )
         */
        if (!tryNoToSurvive) {
            tryNoToSurvive = true;
            log("SURVIVE: try to explore -> EXPLORE");
            state = PigletState.EXPLORE;
            return ;
        }

        tryNoToSurvive = false;

        /*
         	      
         */
        if (E > M * C_crit) {
            log("SURVIVE: has energy");
            return ;
        }

        /*
         	    ,     ( - )
         */
        final float count = pointInfo.getCount(this);
        final float neededToEat = M * C_norm - E;
        log("SURVIVE: eat " + neededToEat + " / " + count);
        throw new EventException(EventKind.ACTION_EAT, neededToEat);
    }

    private void onAttack() throws EventException {
        /*
         	     
         */
        Location enemy = findClosestEnemy();
        float dist = enemy != null ? distance(location, enemy) : Herder.INF;
        if (enemy != null && dist < S + Herder.EPS) {
            if (dist < Herder.EPS) {
                log("ATTACK: attack at " + enemy);
                throw new EventException(EventKind.ACTION_ATTACK, enemy);
            }

            log("ATTACK: move&attack at " + enemy);
            throw new EventException(EventKind.ACTION_MOVE_ATTACK, enemy);
        }

        /*
         	  .  
         */
        log("ATTACK: enemy position lost -> EXPLORE");
        state = PigletState.EXPLORE;
    }

    private boolean isGrowingPeriod() {
        return herder.piglets.size() < C_growth_barrier;
    }

    private DrawWell findClosestDrawWell() {
        herder.lockRead();

        DrawWell found = null;
        float minWeight = Herder.INF;
        
        for (DrawWell dw : herder.drawWells) {
            GladeCell cell = dw.getCell();
            
            // first draw well should be good to born
            if ((cell.count + C_steps_to_born * cell.growth) / (dw.owners.size() + 1) < (M - E) / C_steps_to_born) {
                continue;
            }

            // calc draw well weight
            final float dist = distance(location, dw.pos);

            if (dist > C_max_steps_to_dw * S0) {
                continue;
            }

            float weight = dist - (cell.count + cell.growth * C_growt_weight) * C_energy_weight / dw.owners.size();
            if (weight < minWeight) {
                minWeight = weight;
                found = dw;
            }
        }

        herder.unlockRead();

        return found;
    }

    private boolean checkDistanceConstraints(Location next) {
        // distance from current position
        final float dist = distance(location, next);

        final float turnCount = (float)Math.ceil(dist / S);

        // energy at next (consider to move cost and mass-live cost)
        final float E_at_next = E - dist * Constants.K_movecost
                - turnCount * Constants.K_masscost * M;
        
        return E_at_next > (herder.MAX_TURNS - herder.getTurn() - turnCount) * M * Constants.K_masscost + M * C_crit;
    }

    private Location findClosestUnexploredCell() {
        if (herder.everythingExplored) {
            return null;
        }

        herder.lockRead();

        Location found = null;
        float minDist = Herder.INF;

        boolean hasUnexploredColumnt = false;
        for (int i=0; i<herder.glade.length; ++i) {
            if (herder.exploredColumn[i]) {
                continue;
            }
            hasUnexploredColumnt = true;

            boolean hasUnexplored = false;
            for (int j=0; j<herder.glade[i].length; ++j) {
                if (!herder.glade[i][j].explored) {
                    hasUnexplored = true;
                }

                // check if cell hasn't explored
                if (!herder.glade[i][j].explored && herder.glade[i][j].shouldBeExplored != herder.getTurn()) {
                    Location curPos = new Location(i, j);
                    float dist = distance(location, curPos);
                    if (dist < minDist) {
                        found = curPos;
                        minDist = dist;
                    }
                }
            }

            if (!hasUnexplored) {
                herder.unlockRead();
                herder.lockWrite();
                herder.exploredColumn[i] = true;
                herder.unlockWrite();
                herder.lockRead();
            }
        }

        herder.unlockRead();
        
        // mark glade as "should ne explored"
        if (found != null) {
            herder.lockWrite();
            herder.glade[found.getX()][found.getY()].shouldBeExplored = herder.getTurn();
            herder.unlockWrite();
        }

        if (!hasUnexploredColumnt) {
            herder.lockWrite();
            herder.everythingExplored = true;
            herder.unlockWrite();
        }

        return found;
    }

    private static final int [] EXPLORE_DELTA = new int [] { 0, 1, -1, 2, -2, 3, -3 };
    private Location findExploreStep() {
        herder.lockRead();
        
        final int VARIATION_LENGTH = 3;

        int max = 0;
        Location best = null;

        for (int ED_i : EXPLORE_DELTA) {
            for (int ED_j : EXPLORE_DELTA) {
                // let's calc cells amount that will be explored
                int unexplored = 0;

                for (int di = 0; di < VARIATION_LENGTH; ++di) {
                    for (int dj = 0; dj < VARIATION_LENGTH; ++dj) {
                        int x = normalizeX(location.getX() + ED_i + EXPLORE_DELTA[di]);
                        int y = normalizeY(location.getY() + ED_j + EXPLORE_DELTA[dj]);

                        if (!herder.glade[x][y].explored && herder.glade[x][y].shouldBeExplored != herder.getTurn()) {
                            ++unexplored;
                        }
                    }
                }

                if (max < unexplored || max == unexplored && Util.frnd() < 0.2f) {
                    max = unexplored;
                    best = new Location(
                            normalizeX(location.getX() + ED_i),
                            normalizeY(location.getY() + ED_j));
                }
            }
        }

        herder.unlockRead();
        
        if (best != null) {
            herder.lockWrite();

            // mark cells that it will be explored
            for (int di = 0; di < VARIATION_LENGTH; ++di) {
                for (int dj = 0; dj < VARIATION_LENGTH; ++dj) {
                    int x = normalizeX(best.getX() + EXPLORE_DELTA[di]);
                    int y = normalizeY(best.getY() + EXPLORE_DELTA[dj]);

                    herder.glade[x][y].shouldBeExplored = herder.getTurn();
                }
            }

            herder.unlockWrite();
        }
        
        return best;
    }

    private Location findClosestEnemy() {
        if (herder.totalMass < C_fight_barrier) {
            return null;
        }

        herder.lockRead();

        float minDist = Herder.INF;
        Location found = null;

        for (int ED_i : EXPLORE_DELTA) {
            for (int ED_j : EXPLORE_DELTA) {
                final int x = normalizeX(location.getX() + ED_i);
                final int y = normalizeY(location.getY() + ED_j);

                final Location pos = new Location(x, y); 
                final float dist = distance(location, pos);
                if (herder.glade[x][y].enemy >= herder.getTurn() - C_enemy_history && dist < minDist) {
                    minDist = dist;
                    found = pos;
                }
            }
        }

        herder.unlockRead();
        return found;
    }

    private static final int [] FEEDING_DELTA = new int [] { 0, 1, -1, 2, -2, 3, -3, 4, -4, 5, -5, 6, -6 };
    private Location findClosestFeeding() {
        herder.lockRead();

        Location best = null;
        float maxCount = Herder.EPS;

        for (int FD_i : FEEDING_DELTA) {
            for (int FD_j : FEEDING_DELTA) {
                int x = normalizeX(location.getX() + FD_i);
                int y = normalizeY(location.getY() + FD_j);

                GladeCell cell = herder.glade[x][y];

                if (cell.count > maxCount && cell.growth < Herder.EPS) {
                    maxCount = cell.count;
                    best = new Location(x, y);
                }
            }
        }

        herder.unlockRead();

        return best;
    }

    private boolean isCreated = false;
    public BeingParams getParams() {
        if (!isCreated || isGrowingPeriod()) {
            return new BeingParams(M0, S0);
        }
        return new BeingParams(M * Constants.K_maxbornvariation, S0);
    }

    public void reinit(UserGameInfo userGameInfo) {
        super.reinit(userGameInfo);
    }

    private static final Herder herder = new Herder();
    protected SuperHerder getHerder() {
        return herder;
    }

    public void processEvent(BeingInterface bi, Event event) {
        super.processEvent(bi, event);

        switch (event.kind()) {
            case BEING_BORN:
                isCreated = true;
                break;
        }
    }

    public void log(String s) {
        if (beingInterface != null) {
//            beingInterface.log(this, "(" + E + "/" + M + ") " + s);
        }
    }
}
