package universum.engine;

import java.lang.reflect.Method;
import java.util.*;
import java.util.concurrent.*;
import java.security.*;
import java.security.cert.Certificate;

import universum.bi.*;
import universum.util.*;
import universum.ui.GameOwner;
import universum.ui.GameResult;

import static universum.bi.Constants.*;
import static universum.engine.RemoteProto.*;
/**
 *
 * @author nike
 */
public class Universe 
    implements Runnable, EventListener, EventFilter, RenderingInfo 
{
    private GameOwner controller;
    private Topology topo;
    private ResourceControl rc;
    private EventQueue eq;
    private List<Entity> entities;
    private Map<String,ScoreData> scores;
    private Map<EventFilter,List<EventListener>> listeners;
    private Map<Entity,EntityInfo> info;
    private Map<Location,List<PointInfo>> neighbours;
    private Map<Location,PointInfo> points;
    private Map<Integer, EntityInfo> mapById;
    private Map<PointInfo, Integer> marks;
    //private BeingLoader loader;
    private boolean seventhSeal;
    private boolean paused = false, started = false, stopWhenOneLeft;
    private Object pauser;
    private AccessControlContext creatureContext;
    private BeingInterfaceImpl iface;
    private RemoteRenderer broadcaster;
    private TurnService turnMaker;
    private TurnRecorder recorder;
    private GameKind gameKind;
    private int status;
    private GameInfo gameInfo;

    public Universe(GameOwner controller, GameInfo gameInfo) {
        setStatus(STATUS_UNKNOWN);

        // not so flexible for now
        this.topo = new DefaultTopology(gameInfo.fieldWidth, 
                                        gameInfo.fieldHeight);
        this.rc = new FairResourceControl(gameInfo.numRegular, 
                                          gameInfo.numGolden);
        this.controller = controller;
        this.gameKind = gameInfo.gameKind;
        this.maxTurns = gameInfo.maxTurns;
        this.recording = gameInfo.recordGameFilePath != null;
        // we keep reference to receive updated values
        this.gameInfo = gameInfo;

        seventhSeal = false;
        
        eq = new EventQueue();
        entities = new ArrayList<Entity>();
        proc = new Processor();
        listeners = new HashMap<EventFilter,List<EventListener>>();
        info = new HashMap<Entity,EntityInfo>();
        points = new HashMap<Location,PointInfo>();
        neighbours = new HashMap<Location,List<PointInfo>>();
        scores = new HashMap<String,ScoreData>();
        mapById = new HashMap<Integer, EntityInfo>();
        marks = new HashMap<PointInfo, Integer>();
        pauser = new Object();
        
        ScoreData.reset();

        PermissionCollection perms = makePerms();
        creatureContext = getAccessControlContext(perms);
                
        turnMaker = new TurnService(this);
        
        recorder = recording ? 
            new TurnRecorder(gameInfo.recordGameFilePath) : null;

        topo.init(this);
        rc.init(this);
        
        // make sure we'll get all events we care about
        registerListener(this, this);               

        iface = new BeingInterfaceImpl();

        int port = controller.getListenPort();
        if (port >= 0 || recording) {
            broadcaster = new RemoteRenderer(this, controller);            
        }
        setStatus(STATUS_NOT_STARTED);
        // do some preloading
        System.out.flush();
        if (gameInfo.out != null) {
            gameInfo.out.flush();
        }
    }

    private PermissionCollection makePerms() {       
        return null;     
    }

    private AccessControlContext getAccessControlContext(PermissionCollection perms) {       
        ProtectionDomain domain =
            new ProtectionDomain(new CodeSource(null,
                                                (Certificate[]) null), perms);        
        AccessControlContext acc =
            new AccessControlContext(new ProtectionDomain[] { domain });
        
        return acc;
    }
   
    
    public void setController(GameOwner controller) {
        this.controller = controller;
    }
    
    public boolean addZombie(Entity e) {
        synchronized (zombies) {
            return zombies.add(e);
        }
    }

    public boolean isZombie(Entity e) {
         synchronized (zombies) {
            return zombies.contains(e);
        }
    }

    private void death(EntityInfo ei) {
        Being b = ei.isBeing ? (Being)ei.entity : null;
        if (b == null) return;

        // if already dead - do nothing
        if (!addZombie(b)) {
            return;
        }
	if (getDebug()) {
	    System.out.println(ei.name+" of "+ei.player+
			       "["+ei.id+"] just died");
	}
        sendEvent(new Event(EventKind.BEING_DEAD, ei.id, null), null);        
    }
    private Set<Entity> zombies = new HashSet<Entity>();
   
    // RenderingInfo implementation
    public  Integer getId(Entity e) {
        return iface.getId(e);
    }

    public Location getLocation(Integer id) {
	EntityInfo ei = getInfoById(id);
	if (ei == null) {
	    return null;
	}
        return ei.location();
    }
    
    public float getEnergy(Integer id) {
	EntityInfo ei = getInfoById(id);
	if (ei == null) {
	    return 0f;
	}
        return ei.energy;
    }

    public String getOwner(Integer id) {
        EntityInfo ei = getInfoById(id);
        if (ei == null) {
	    return null;
	}	
        return ei.player;
    }

    public int getOwnerId(Integer id) {
        EntityInfo ei = getInfoById(id);
        if (ei == null) {
	    return -1;
	}	
        return scores.get(ei.player).id;
    }
    
    public float getMass(Integer id) {
        EntityInfo ei = getInfoById(id);
        if (ei == null) return 0f;
        
        return ei.params.M;
    }
    
    public float getResourceCount(Location loc) {
        return getResourceControl().getCount(loc);
    }
    
    public List<Integer> diameters() {
        return getTopo().diameters();
    }

    public int numTurns() {
        return this.numTurns;
    }

    public int maxTurns() {
        return this.maxTurns;
    }   
    
    public void forAllE(final Walker<Integer> what) {
	forAllEntities(new Walker<Entity>() {
	    public void walk(Entity e) {
		Integer id = getId(e);
		what.walk(id);
	    }
	});
    }

    public void forAllR(Walker<Location> what) {
        getResourceControl().forAllResourceLocations(what);
    }

    public synchronized int getStatus() {
        return this.status;
    }

    public Map<String,ScoreData> scores() {
        return this.scores;
    }

    public String nameById(int num) {
        synchronized (scores) {
            if (num < currentResults.size()) {
                return currentResults.get(num).player;
            } else {
                return null;
            }
        }
    }

    //// end of RenderingInfo impl

    public synchronized void setStatus(int status) {
        if (this.status >= STATUS_COMPLETED && status < this.status) {
            return;
        }
        this.status = status;
    }

    class BeingInterfaceImpl implements BeingInterface {
        BeingInterfaceImpl() {}
        
        // of EntityInterface
        public Location getLocation(Entity b) {
            EntityInfo ei = getInfo(b);
            if (ei == null) {
                return null;
            }
            return  ei.location();
        }
        // of BeingInterface
        public Integer getId(Entity e) {
            EntityInfo ei = getInfo(e);
            if (ei == null) {
                return null;
            }
            return ei.id;
        }
    
        public float distance(Location one, Location another) {
            return getTopo().distance(one, another);
        }
    
        public synchronized void log(Being b, String s) {
            if (gameInfo.out != null) {
                EntityInfo ei = b != null ? getInfo(b) : null;
                gameInfo.out.println((ei != null ? ei.player + "[" +ei.id+"]: " : "" )+s);
            }
        }  

        public float getEnergy(Being b) {
            EntityInfo ei = getInfo(b);
            if (ei == null) {
                return 0f;
            }
            assert ei.isBeing;
            return  ei.energy();
        }
            
        public PointInfo getPointInfo(Being b) {
            EntityInfo ei = getInfo(b);
            if (ei == null) {
                return null;
            }
            assert ei.isBeing;
            return getPointInfo(ei.location());
        }
    
        private PointInfo getPointInfo(final Location loc) {
            synchronized (points) {
                // we lazily allocate PointInfos
                PointInfo pi = points.get(loc);
                if (pi == null) {
                    pi = AccessController.doPrivileged(new PrivilegedAction<PointInfo>() {
                            public PointInfo run() {
                                return new PointInfo(Universe.this, loc);
                            }
                            });
                    points.put(loc, pi);
                }
                return pi;
            }
        }
    
        public List<PointInfo> getNeighbourInfo(Being b) {
            EntityInfo ei = getInfo(b);
            assert ei != null && ei.isBeing;
            Location loc = ei.location();
            List<PointInfo> rv;
            synchronized (neighbours) {
                rv = neighbours.get(loc);
                if (rv == null) {
                    List<Location> locations = topo.getNeighbours(loc);
                    int lc = locations.size();
                    PointInfo[] points = new PointInfo[lc];
                    for (int i = 0; i < lc; i++) {
                        points[i] = getPointInfo(locations.get(i));
                    }
                    rv = new ImmutableList<PointInfo>(points);
                    neighbours.put(loc, rv);
                }
            }
            return rv;
        }

        public List<Location> getReachableLocations(Being b) {
            final EntityInfo ei = getInfo(b);
            assert ei != null && ei.isBeing;
            
            return topo.getNeighbours(ei.location(), ei.params.S);
        }       
    
        public Object getOwner(Being me, Integer id) {
            EntityInfo ei = getInfoById(id);
            if (ei == null) return null;
        
            //return checkDistance(ei, getInfo(me)) ? ei.player : null;
            return ei.playerHandle;
        }

        public float getMass(Being me, Integer id) {
            EntityInfo ei = getInfoById(id);
            if (ei == null) return 0f;
        
            return checkDistance(ei, getInfo(me)) ? ei.params.M : 0f;
        }
    }


    abstract class UserTask<T> implements  PrivilegedAction<T>,
					   Comparable<UserTask> {	
	long started;
	Entity e;
	UserTask(Entity e) {
	    this.e = e;
	}

	abstract protected T callUser();
	
	public T run() {
	    started = System.currentTimeMillis();
            return callUser();
        }
	
	public int compareTo(UserTask other) {
	    long thisVal = this.started;
	    long otherVal = other.started;
	    return (thisVal < otherVal ? -1 : (thisVal == otherVal ? 0 : 1));
	}
    }

    class  TurnTask extends UserTask<Event> {
	TurnTask(Entity e) {
	    super(e);
	}
	// this code is called from user context
	protected Event callUser() {
	    return e.makeTurn(iface);
	}
		
    }        

    // we use here Callable, not Runnable, to avoid name clash with PrivilegedAction's
    // run(), otherwise Runnable would be enough
    // as side effect - we may know when event was processed
    class  EventHandleTask extends UserTask<Object> implements Callable<Object> {
	Event ev;
	EventHandleTask(Entity dest, Event ev) {
	    super(dest);
	    this.ev = ev;
	}
	// this code is called from user context
	protected Object callUser() {
	    e.processEvent(iface, ev);
	    return null;
	}

	// this code is called from worker thread in engine context
	public Object call() {
	    proc.addCurrent(this);
	    try {
		invokeInSecureContext(this);
	    } catch (Throwable t) {
                if (t instanceof ThreadDeath) {
                    System.out.println("worker interrputed (event)");   
                } else {
                    t.printStackTrace();
                }
                addZombie(e);
            } finally {
		proc.removeCurrent(this);
	    }
	    return null;
	}
    }
    

    class Processor implements Walker<Entity> {
        class BornRequest {
            float       energy;
            BeingParams bp;
            EntityInfo  parent;
            Class       klazz;
            Location    location;
            BornRequest(EntityInfo parent, Location location, 
                        BeingParams bp, float energy) {
                this.energy = energy;
                this.bp = bp;
                this.parent = parent;
                this.location = location;
            }
        }

	// somewhat hard to read code, but idea is to feed tasks to
	// multiple threads. idx just works to reference to next
	// to ne executed task
        class EntityFeeder implements Feeder<Entity>, Runnable {
            int idx, size;
            Walker<Entity> walker;

            EntityFeeder(Walker<Entity> walker) {
                this.walker = walker;
            }

            void reset() {
                idx = 0; size = entities.size();
            }
            public synchronized Entity next() {
		if (seventhSeal) {
		    return null;
		}
                assert size == entities.size();
                if (idx < size) {
                    return entities.get(idx++);
                }
                return null;
            }
            
            public void run() {                
                Entity e;
                while ((e = next()) != null) {
                    walker.walk(e);
                }
            }
        }        

        private List<BornRequest> babies = new LinkedList<BornRequest>();
        private EntityFeeder turnFeeder = new EntityFeeder(Processor.this);
        private HashMap<UserTask,Thread> current = new HashMap<UserTask,Thread>();
        
        void proceedTurn() {           
            // delegate to TurnService
            turnFeeder.reset();
            // schedule calls to makeTurn() of all creatures
            // by invoking Processor.walk() from several worker threads
            turnMaker.executeMulti(turnFeeder);
        }

        void proceedEvent(Entity dest, Event e) {
	    // we use TurnService to request execute of user code
	    turnMaker.executeSingle(new EventHandleTask(dest, e));       
        }

        private void addCurrent(UserTask wt) {
            synchronized (current) {
                current.put(wt, Thread.currentThread());
            }
        }

        private void removeCurrent(UserTask wt) {
            synchronized (current) {
                current.remove(wt);
            }
        }

        public @SuppressWarnings("deprecation") void killCurrent() {
            synchronized (current) {
		long now = System.currentTimeMillis();
		// removal from hashtable while iterating not exactly a good idea
		Vector<UserTask> zut = new Vector<UserTask>();
                for (UserTask wt : current.keySet()) {
		    if (now - wt.started > 1000) {
			Thread t = current.get(wt);
			assert t != Thread.currentThread();
			t.stop();
			// it will be added to zombies by interrupted thread 
			// but do it here just in case
			addZombie(wt.e);
			zut.add(wt);
		    }
                }
		
		for (UserTask z : zut) {
		    removeCurrent(z);
		}
            }
        }	

        // be careful - this routine is invoked by multiple threads
        public void walk(final Entity e) {            
            EntityInfo ei = getInfo(e);
            if (ei == null) return;

            // let it make turn and send event if it made one and able to
            Event ev;
	    TurnTask tt = new TurnTask(e);
            addCurrent(tt);
            try {        
                ev = invokeInSecureContext(tt);
            } catch (Throwable t) {                
                if (t instanceof ThreadDeath) {
                    System.out.println("worker interrputed (turn)");
                } else {
                    t.printStackTrace();
                }
                addZombie(e);
                return;
            } finally {
                removeCurrent(tt);
            }
            // now update energy level according to creature parameters
            updateEntity(ei, ev);
        }


        void handleEvent(Entity dest, Event e) {
	    EventHandleTask eht = new EventHandleTask(dest, e);
            addCurrent(eht);
	    try {
		invokeInSecureContext(eht);
	    } catch (Throwable t) {
                if (t instanceof ThreadDeath) {
                    System.out.println("worker interrputed (event)");   
                } else {
                    t.printStackTrace();
                }
                addZombie(dest);
            } finally {
		removeCurrent(eht);
	    }
        }

        private void updateEntity(EntityInfo ei, Event ev) {
            if (!ei.isBeing) {
                return;
            }
            float dE = eventCost(ev, ei);
            if (dE >= 0f) {
                // if creature can do and got enough energy to do what it planned - let it do it
                if (ev != null && dE < ei.energy()) {
                    if (getDebug()) {
                        debug(ei.entity, ev, true);
                    }
                    sendEvent(ev, ei.id);
                }
            } else {
                dE = 0f;
            }
            
            // now add living cost and update energy level
            dE += ei.eMass * K_masscost;                        
            
            ei.energyDelta(-dE);
        }
        
        final void postiterate() {
            drainQueue();
            
            // at this point no other mutators of zombies, but just in case
            synchronized (zombies) {
                for (Entity e : zombies) {
                    removeEntity(e, true);
                }
                zombies.clear();
            }
            
            synchronized (babies) {
                for (BornRequest br : babies) {
                    addBeing(br.parent, br.bp, br.energy, br.location);
                }
                babies.clear();
            }
        }      

        private float eventCost(Event ev, EntityInfo ei) {
            if (ev == null) {
                return 0f;
            }
            // handle all known cases, ignore others
            switch (ev.kind()) {
                case ACTION_MOVE_TO:
                {
                    Object o = ev.param();
                    if ((o == null ) || !(o instanceof Location)) {
                        return -1f;
                    }
                    return maybeMove(ei, (Location)o);
                }
                
                case ACTION_BORN:
                {
                    Object o = ev.param();
                    if ((o == null ) || !(o instanceof BeingParams)) {
                        return -1f;
                    }
                    BeingParams newbp = ((BeingParams)o).clone();
                    return maybeBorn(ei, ei.params, newbp);
                }
                
                case ACTION_ATTACK:
                {
                    Object o = ev.param();
                    if ((gameKind == GameKind.PEACE_DUEL)  || 
                        (o == null ) || !(o instanceof Integer)) {
                        return -1f;
                    }
                    return maybeAttack(ei, (Integer)o);
                }
                
                case ACTION_MOVE_ATTACK:
                {
                    Object o = ev.param();
                    if ((gameKind == GameKind.PEACE_DUEL) ||
                        (o == null ) || !(o instanceof Integer)) {
                        return -1f;
                    }
                    return maybeMoveAttack(ei, (Integer)o);
                }
                
	        case ACTION_EAT: 
		{
		    Object o = ev.param();
                    if ((o == null ) || !(o instanceof Float)) {
                        return -1f;
                    }
                    return maybeEat(ei, (Float)o);
		}
		
		case ACTION_GIVE:
		{
		    Integer target = ev.target();
		    Object o = ev.param();
                    if ((o == null ) || !(o instanceof Float)) {
                        return -1f;
                    }
                    return maybeGive(ei, target, (Float)o);
		}

		/*
                case ACTION_MARK: 
                {
                    Object o = ev.param();
                    if (o == null) {
                        return -1f;
                    }
                    return maybeMark(ei, o);
		    } */

                default:
                {
                    return -1f;
                }                
            }
        }
        
        private boolean checkBorn(float o, float n) {
            return 
                (n >= K_minbornvariation * o) && (n <= K_maxbornvariation * o);
        }
        
        private synchronized float maybeBorn(EntityInfo ei, 
                                             BeingParams mybp, 
                                             BeingParams newbp) {
            // if requested params are too different
            if (!checkBorn(mybp.M, newbp.M) || !checkBorn(mybp.S, newbp.S)) {
                return -1f;
            }
            float energy = ei.energy();

            if (energy < K_toborn * mybp.M) {
                return -1f;
            }

            Location where = ei.location();
            if (K_maxmassperpoint > 0f) {                            
                if (totalMass(where) + newbp.M > K_maxmassperpoint) {
                    return -1f;
                }
            }
                        
            float babyE = energy/2f; // baby gets half of parent's energy
            synchronized (babies) {
                babies.add(new BornRequest(ei, where, newbp, babyE));
            }
            
            return babyE + mybp.M * K_borncost;
        }
    }

    private synchronized void debug(Entity e, Event ev, boolean out) {
        iface.log((Being)e, (out ? ">>>" : "<<<")+ev);
    }

    // synchronized to avoid potential race on consumption
    private synchronized float consumeEnergy(EntityInfo ei, float count) {
        assert ei != null && ei.isBeing;

        // be sane
        if (count <= 0f) return 0f;
       
        // limit single bite
        if (count > ei.params.M * K_bite) {
            count = ei.params.M * K_bite;
        }
        
        float can = getResourceControl().consume(ei.location(), count);
        float did = ei.energyDelta(can);
        assert did <= can;        
        // give back something we cannot in fact eat up
        getResourceControl().consume(ei.location(), did-can);

        return did;
    }        
    
    private float giveEnergyTo(final EntityInfo ei, final Integer other, 
                               float count) {
        assert ei != null && ei.isBeing;
        // trying to cheat - attack by giving
        if (count <= 0f) {
            return 0f;
        }
        
        if (count > ei.energy()) {
            count = ei.energy();
        }
        EntityInfo acceptor = getInfoById(other);
        // if trying to give to non-being, or someone we're not on the same cell
        if (!acceptor.isBeing || !checkDistance(acceptor, ei)) {
            return 0f;
        }
        // cannot give to already dead ones
        if (isZombie(acceptor.entity)) {
            return 0f;
        }
        
        ei.energyDelta(-count);
        acceptor.energyDelta(count);
        
        sendEvent(new Event(EventKind.BEING_ENERGY_GIVEN, other, 
                            new Float(count)),
                  ei.id);
        
        return count;
    }

    private float totalMass(Location l) {
        PointInfo pi = iface.getPointInfo(l);
        return pi != null ? pi.totalMass(null) : 0f;
    }


    private boolean checkDistance(EntityInfo one, EntityInfo other) {
        return one != null && other != null && 
            one.location().equals(other.location());
    }
    
    public boolean checkDistance(Entity one, Location loc) {
        if (one == null || loc == null) return false;
        Location bloc = iface.getLocation(one);
        return (getTopo().distance(bloc, loc) <= 1f);
    }
    
    private EntityInfo getInfo(Entity e) {
        synchronized (info) {
            EntityInfo ei =  info.get(e);           
            return ei;
        }
    }

    private void removeInfo(Entity e) {
        synchronized (info) {
            info.remove(e);
        }
    }
    
    private EntityInfo getInfoById(Integer id) {
        synchronized (mapById) {
            return  mapById.get(id);
        }
    }

    private void drainQueue() {
        Event e;
        while ((e = eq.getEvent()) != null) {
            processEvent(e);
        }
        // wait for completion of event processing
        waitCompletion();
    }
        
    private Processor proc;    
    
    public void bigbang() {
        new Thread(this).start();
        numTurns = 0;        
    }
    
    private void waitCompletion() {
        turnMaker.waitCompletion();       
    }

    public void run() {
        started = true;
        setStatus(STATUS_RUNNING);
        recordState();        
        controller.redraw();

        // have to be after state recording, so that clients can see state 
        // before start
        Util.sleep(gameInfo.waitBefore);

        while (!seventhSeal)  {
            // needed, to allow INIT events to be dispatched, before event processing
            drainQueue();            

            // let beings perform their actions
            if (!paused()) {
                synchronized (entities) {
                    proc.proceedTurn();
                }
                // now wait for completion, not under lock
                
                waitCompletion();

                // kill all candidates for death, give birth to all who
                // should be born
                proc.postiterate();
                
                drainQueue();                               
            }
                       
            // update resource counters
            getResourceControl().makeTurn();
            
            // compute player's score
            updateScores();
            
            // update marks
            updateMarks();

            turnEnd();
        }

        setStatus(STATUS_FINISHED);
        
        // make an official apocalypse end
        seventhSeal = false;
    }    
    
    class ScoreCounter implements Walker<Entity> {
        public void walk(Entity e) {
            if (!(e instanceof Being)) return; // who cares about stones
            Being b = (Being)e;
            EntityInfo ei = getInfo(b);           
            if (ei != null) {
                String player = ei.player;
                ScoreData sd = scores.get(player);
                assert sd != null;
                sd.score += ei.params.M;
                sd.energy += ei.energy();
                sd.alive = true;
            }
        }
    }
    ScoreCounter scoreCounter = new ScoreCounter();
    List<ScoreData> currentResults = new ArrayList<ScoreData>();
    private void updateScores() {
        synchronized (scores) {
            // preclean
            for (ScoreData sd : scores.values()) {
                sd.score = 0; sd.energy = 0; sd.alive = false;
            }
            forAllEntities(scoreCounter);

            currentResults.clear();
            currentResults.addAll(scores.values());
            Collections.sort(currentResults);
        }
    }
   
    private void updateMarks() {
      /*
      for (Iterator<PointInfo> i = marks.keySet().iterator(); i.hasNext();) {
            PointInfo pi = i.next();
            Integer ttl = marks.get(pi);
            if (ttl.intValue() < 1) {
                pi.putMark(null);
                i.remove();
            } else {
                marks.put(pi, ttl-1);
            }
        }
      */
    }

    /****************** UI *********************/
    public boolean started() {
        return this.started;
    }
    
    public boolean paused() {
        return this.paused;
    }
    
    public void setPaused(boolean paused) {
        synchronized (pauser) {            
            this.paused = paused;
            pauser.notifyAll();
        }
    }
   
    /******************************************/
    
    public synchronized void registerListener(EventListener listener, EventFilter filter) {
        synchronized (listeners) {
            List<EventListener> list = listeners.get(filter);
            if (list == null) {
                list = new ArrayList<EventListener>();
                listeners.put(filter, list);
            }
            // we don't use sets to keep ordering of listeners (XXX?)
            if (!list.contains(listener)) {
                list.add(listener);
            }
        }
    }
    
     public synchronized void removeListener(EventListener listener, EventFilter filter) {
         synchronized (listeners) {
             List<EventListener> list = listeners.get(filter);
             if (list == null) {
                 return;
             }
             list.remove(listener);
         }
     }
    
    public void sendEvent(Event e, Integer source) {
        if (e != null) {
            if (source != null) {
                ((BasicEvent)e).setSender(source);
            }
            eq.putEvent(e);
        }
    }
    
    private void killemall() {
        synchronized(entities) {
            entities.clear();
        }
    }
    
    private void removeEntity(Entity e, boolean natural) {
        synchronized(entities) {
            EntityInfo ei = getInfo(e);
            // already removed by other mean
            if (ei == null) {
                return;
            }
            // put dead corpse on the ground
            if (natural && e instanceof Being && ei.energy() > 0f) {
                ResourceControl rc = getResourceControl();
                rc.addSource(ei.location, ei.energy(), 0f, -1f);
            }
            entities.remove(e);
            // we use entities object as lock for operations
            // on global lists
            removeInfo(e);
        
            // remove it from the map
            PointInfo pi = iface.getPointInfo(ei.location);
            pi.removeEntity(ei.id);
    
            synchronized (mapById) {
                mapById.remove(ei.id);
            }
            
            removeListener(ei, ei);
        }
    }     
    
    public void apocalypse() {
	seventhSeal = true;
        if (broadcaster != null) {
            broadcaster.stop();
        }
	if (recorder != null && recording) {
	    recorder.finishRecording();
	}
	turnMaker.stop();	
        //killemall();
        controller.redraw();        
        started = false;
    }
    
    // of EventListener
    public void eventFired(Event e) {
       // we ignore messages from beings
       // we know about all events, but doing nothing for now
    }
        
    // of EventFilter
    public boolean match(BasicEvent e) {
        // for now we'll look at all events
        return true;
    }

    private float maybeAttack(EntityInfo ei, Integer other) {
        assert ei != null && ei.isBeing;
        
        EntityInfo oi = getInfoById(other);
        if (!checkDistance(oi, ei)) {
            return 0f;
        }
        
        // shouldn't attack dead ones
        if (isZombie(oi.entity)) {
            return 0f;
        }
        

        float cost = ei.params.M * K_fightcost;
        if (cost > ei.energy) {
            return -1f;
        }

        float dE = oi.energyDelta(-ei.params.M * K_fight);       
        sendEvent(new Event(EventKind.BEING_ATTACKED, other, new Float(-dE)), ei.id);
        
        return cost + K_retaliate * oi.params.M;
    }
    
    private float maybeMove(EntityInfo ei, Location newLoc) {
        float distance = getTopo().distance(ei.location, newLoc);
        float speed = ei.params.S;
        if (distance > speed) {
            System.out.println("speed is not enough: "+speed+"<"+distance);
            return -1f;
        }
        float m = totalMass(newLoc);
        if (K_maxmassperpoint > 0f && m + ei.params.M > K_maxmassperpoint) {
            System.out.println("cannot move in: too crowded");
            return -1f;
        }
        
        ei.setLocation(newLoc);
        return speed * K_movecost;
    }
    
    private float maybeMoveAttack(EntityInfo ei, Integer other) {
        assert ei != null && ei.isBeing;
        
        EntityInfo oi =  getInfoById(other);
        
        if (oi == null) return -1f;     
        
        // if out of reach
        if (getTopo().distance(oi.location(), ei.location()) > ei.params.S) {
            return -1f;
        }
                        
               
        float cost = ei.params.M * K_fightcost + ei.params.S*K_movecost;
        if (cost > ei.energy) {
            return -1f;
        }

        Location newLoc = oi.location();
        float m = totalMass(newLoc);
        if (K_maxmassperpoint < 0f || m + ei.params.M <= K_maxmassperpoint) {
            ei.setLocation(newLoc);
        }
        
        float dE = oi.energyDelta(-ei.params.M * K_fight * K_fightmovepenalty);                        
        sendEvent(new Event(EventKind.BEING_ATTACKED, other, new Float(-dE)), ei.id);
        
        return cost + K_retaliate * oi.params.M;
    }
    

    private float maybeEat(EntityInfo ei, Float amount) {
	consumeEnergy(ei, amount);
	// eating has no cost
	return 0f;
    }

    private float maybeGive(EntityInfo ei, Integer who, Float amount) {
	giveEnergyTo(ei, who, amount);
	// giving has no cost
	return 0f;
    }

    /*
    private float maybeMark(EntityInfo ei, Object mark) {
        PointInfo pi = iface.getPointInfo(ei.location());
        if (pi.getMark() != null) {
            return -1f;
        }
        pi.putMark(mark);
        marks.put(pi, new Integer(K_markttl));
	return K_markcost;
	}*/
    
    public void forAllEntities(Walker<Entity> walker) {
        synchronized (entities) {
            Util.walkList(entities, walker);
        }
    }
    
    public void processEvent(Event e) {
        synchronized (listeners) {
            for (EventFilter f : listeners.keySet()) {
                if (f.match(e)) {
                    // if someone throws an exception - all other listeners
                    // will miss the notification
                    try {
                        for (EventListener l : listeners.get(f)) {
                            l.eventFired(e);
                        }
                    } catch (InterruptedWalkingException iwe) {}
                }
            }
        }
    }
    
    private int numTurns = 0;
    private int maxTurns = 0;    
    private boolean recording = false;   

    void recordState() {
        if (broadcaster != null) {
            broadcaster.makeSnapshot(recording);
            if (recording) {
                recorder.record(numTurns,
                                broadcaster.getSnappedDataLen(),
                                broadcaster.getSnappedData());
            }
        }
    }

    static final boolean consistencyCheck = false;
    void turnEnd() {
        if (consistencyCheck) {
            checkConsistency();
        }
        if (paused()) {
            try {
                synchronized (pauser) {
                    pauser.wait();
                    setStatus(paused ? STATUS_PAUSED : STATUS_RUNNING);
                    recordState();
                }
            } catch (InterruptedException e) {
            }
        } else {
            if (numTurns++ % Constants.getRefreshTurns() == 0) {
                controller.redraw();
            }
            recordState();
        }
        
        if (getTurnDelay() > 0) {
            Util.sleep(getTurnDelay());
        }

        if (gameOver()) {            
            GameResult result = new GameResult(scores(), numTurns, gameKind);
            if (broadcaster != null) {
                broadcaster.setResult(result);
            }
            
            setPaused(true);
            setStatus(STATUS_COMPLETED);
            
            recordState();
            if (recording) {
                recorder.finishRecording();
            }

            controller.redraw();
            controller.notifyAboutCompletion(result);
        }
    }

    boolean gameOver() {
        if (numTurns >= maxTurns || !started || seventhSeal) {
            return true;
        }
        int numAlive = 0;
        for (ScoreData sd : scores.values()) {
            if (sd.alive) {
                numAlive++;
            }
        }
        if (gameKind == GameKind.SINGLE) {
            return numAlive == 0;
        }

         if (gameKind == GameKind.DEBUG) {
            return numAlive == 0;
        }
        
        return numAlive < 2;
    }
    
    class ConsistencyChecker implements Walker<Entity> {
        public void walk(Entity e) {
            Location loc = iface.getLocation(e);
            PointInfo pi = iface.getPointInfo(loc);
            EntityInfo ei = getInfo(e);
            
            Integer entities[] = new Integer[1];
            entities = pi.getEntities(e,  entities);
            
            boolean found = false;
            for (Integer o : entities) {
                if (o.intValue() == ei.id.intValue()) {
                    found = true;
                    break;
                }
            }
            
            if (!found) {
                throw new RuntimeException("Inconsistent: "+e+" in "+loc+
                                           " not in pi");
            } 
        }
    }
    ConsistencyChecker checker;
    void checkConsistency() {
        if (checker == null) {
            checker = new ConsistencyChecker();
        }
        forAllEntities(checker);
    }
       
    static class JailedInfo {
        String player, asString, name;
        BeingParams bp;
        JailedInfo(BeingParams bp) {
            this.bp = bp;
        }
    }

    private Being addBeing(EntityInfo parent,
                           BeingParams bp, float energy, 
                           Location loc) {
        return addBeing(parent.entity.getClass(),
                        parent.player,
                        parent.asString,
                        parent.name,
                        parent.jarFile,
                        bp, energy, loc);
                        
    }

    private Being addBeing(final Class klazz,
                           final String owner,
                           final String asString,
                           final String name,
                           final String jarFile,
                           BeingParams bp, float energy, 
                           Location loc) {
        final JailedInfo ji = new JailedInfo(bp);
        
        Being b = invokeInSecureContext(new PrivilegedAction<Being>() {
              public Being run() {
                Being b;
                try {
                    b = (Being)klazz.newInstance();
                    
                    if (ji.bp == null) {
                        ji.bp = b.getParams().clone();
                    }

                    // not yet known - ask creature
                    if (owner == null) {
			UserGameInfo ugi = new UserGameInfo();
			ugi.kind = gameKind;
			ugi.maxTurns = maxTurns;
                        // call reinit before everything else
                        b.reinit(ugi);
                        ji.player = b.getOwnerName();
                        ji.asString = b.toString();
                        ji.name = b.getName();
                    } else {
                        ji.player = owner;
                        ji.asString = asString;
                        ji.name = name;
                    }
                    return b;
                } catch (Throwable t) {
                    t.printStackTrace();
                    return null;
                }
              }});

        if (b == null) {
            return null;
        }       
        
        if (!checkParams(ji.bp)) {
            System.out.println("Bad params of the being, not adding");
            return null;
        }                
        
        if (owner == null) {
            synchronized (scores) {
                ScoreData sd = scores.get(ji.player); 
                int i = 1;
                String player = ji.player;
                while (sd != null) {
                    ji.player = player+"-"+(i++);
                    sd = scores.get(ji.player); 
                }
                scores.put(ji.player, new ScoreData(ji.player, jarFile));
            }
        }

        // create EntityInfo, it'll add itself into global hash automatically
        new EntityInfo(b, ji.bp, ji.player, ji.asString, 
                       ji.name, loc, energy, jarFile);

        return b;
    }
    
    @SuppressWarnings("unchecked") 
    public Being addBeing(final String klazzName, final String jarFile) {
        final BeingLoader loader = new BeingLoader(this);
        Class klazz = 
            invokeInSecureContext(
                     new PrivilegedAction<Class>() {
                         public Class run() {
                             try {
                                 return loader.loadBeing(klazzName, jarFile);
                             } catch (Throwable t) {
                                 t.printStackTrace();
                             }
                             return null;
                         }
                     });
        


        if (klazz == null) {
            return null;
        }
        
        // now check for potentially dangerous overrides
        try {
            Method hc = klazz.getMethod("hashCode");
            if (!java.lang.Object.class.getMethod("hashCode").equals(hc)) {
                System.out.println("It's disallowed to override hashCode()!");
                return null;
            }
            Method eq = klazz.getMethod("equals", Object.class);
            if (!java.lang.Object.class.getMethod("equals", Object.class).equals(eq)) {
                System.out.println("It's disallowed to override equals()!");
                return null;
            }                    
        } catch (Exception e) {
            e.printStackTrace();
            return null;
        }
        Being b = addBeing(klazz, 
                           null /* owner */, 
                           null /* asString */, 
                           null /* name */,
                           jarFile /* jarFile */,
                           null /* params */, 
                           -1f /* energy */, 
                           null /* loc */);        

        // to dispatch pending events
        drainQueue();

        updateScores();

        return b;
    }    

    public String getOwnerOf(Being b) {
        EntityInfo ei = getInfo(b);
        if (ei == null) {
            return null;
        }
        return ei.player;
    }
    
    private boolean checkParams(BeingParams p) {
        if (p.M < K_minmass || p.M > K_maxmass) {
            return false;
            
        }
        
        if (p.S < K_minspeed || p.S > K_maxspeed) {
            return false;
        }
        
        return true;
    }
    
    public Topology getTopo() {
        return topo;
    }
       
    public ResourceControl getResourceControl() {
        return rc;
    }
    
    public GameOwner getController() {
        return controller;
    }

    private int eid;    
    private HashMap<String, Object> handles =  new HashMap<String, Object>();
    private Object getPlayerHandle(String name) {
        synchronized (handles) {
            Object rv = handles.get(name);
            if (rv == null) {
                rv = new Object();
                handles.put(name, rv);
            }
            return rv;
        }
    }

    class EntityInfo implements EventListener, EventFilter {
        final Entity      entity;
        final String      player;
        final String      asString;
        final String      name;
        private Location  location;
        private final String jarFile;
        final Integer     id;
        // in fact most entities expected to be beings, if not -
        // split out to separate class
        final boolean     isBeing;
        // params
        final BeingParams params;
        private float     energy;
        private float     eMass;
        final Object      playerHandle;    

        EntityInfo(Entity entity, BeingParams bp,
                   String player, String asString, String name,
                   Location loc, float energy, String jarFile) {
            this.entity = entity;
            this.isBeing = entity instanceof Being;
            this.player = player;
            this.asString = asString;
            this.name = name;
            this.id = new Integer(++eid);
            this.jarFile = jarFile;
            registerListener(this, this);
            this.params  = bp.clone();
            setLocation(loc != null ? loc : getTopo().getRandomLocation());
            // being starts fully loaded, unless otherwise specified
            this.energy = energy < 0f ? params.M : energy;
            this.eMass = (float)Math.pow(params.M, 0.5 );
            this.playerHandle = getPlayerHandle(player);

            synchronized (info) {
                info.put(entity, this);
            }  

            synchronized (mapById) {
                mapById.put(id, this);
            }
            synchronized(entities) {
                entities.add(entity);
            }
            
            // must be the last call, when everythings else is already inited
            // note that we pass orginal bp, not params, for better isolation
            sendEvent(new Event(EventKind.BEING_BORN, id, bp), null);
        }
        
        synchronized void setLocation(Location loc) {
            // ignore subspace jumps
            if (!getTopo().contains(loc)) return;

            if (location != null) {
                PointInfo piOld = iface.getPointInfo(this.location);
                piOld.removeEntity(id);
            }
            PointInfo piNew = iface.getPointInfo(loc);
            piNew.addEntity(id);
            this.location = loc;
        }
        
        synchronized Location location() { 
            return this.location; 
        }
        
        synchronized float energyDelta(float delta) {
            if (energy <= 0f) return 0f;

            float newEnergy = energy+delta;
            if (newEnergy < 0f) newEnergy = 0f;
            if (newEnergy > params.M) newEnergy = params.M;
            float rv = newEnergy - energy;


            // to handle float rounding errors
            if (delta > 0f && rv > delta) {
                rv = delta;
            }            

            energy = newEnergy;

            if (isBeing && newEnergy <=  K_emin *params.M) {
                death(this);                
            }

            return rv;
        }
              
        synchronized float energy() {
            return this.energy;
        }
        
        public void eventFired(Event e) {
            if (getDebug()) {
                debug(entity, e, false);
            }
            proc.proceedEvent(entity, e);          
        }
            
        public boolean match(BasicEvent e) {
            Integer target = e.target();
            return target == null ? false : id.equals(target);
        }
    }

    <T> T invokeInSecureContext(PrivilegedAction<T> what) {
        if (JungleSecurity.getCheckSecurity()) {
            return AccessController.doPrivileged(what, creatureContext);
        } else {
            return what.run();
        }
    }

    void checkHanging() {
        System.out.println("Hang?");
        proc.killCurrent();
    }

    int getTurnDelay() {
        return gameInfo.turnDelay; 
    }

}
